using Printf

function verify(day, task, case, expresult)
    taskfun = getfield(Main, Symbol("task", task - 1))
    filename = "day$(@sprintf "%02d" day)\\data\\case$(@sprintf "%02d" case).txt"

    println("Day $day – Task $task – Case $case")
    (result = open(taskfun, filename)) |> display
    @assert result == expresult
end

##

function parseinput(file)
    input = (eachline(file) .|>
             (x -> permutedims(split(x, "") .|>
                               (x -> parse(Int, x))))
    ) |>
            x -> vcat(x...)

    padded = fill(maximum(vec(input)) + 1, (size(input) .+ 2))
    padded[2:(end-1), 2:(end-1)] = input

    return padded
end

function findlocmin(mappe)
    mins = []

    for x ∈ 2:(size(mappe, 2)-1), y ∈ 2:(size(mappe, 1)-1)
        if sum(mappe[CartesianIndex.([y, y, y + 1, y - 1], [x + 1, x - 1, x, x])] .> mappe[y, x]) == 4
            push!(mins, to_indices(mappe, (y, x)))
        end
    end


    return CartesianIndex.(mins)
end

##

function task0(file)
    mappe = parseinput(file)

    sum(mappe[findlocmin(mappe)] .+ 1)
end

##

verify(9, 1, 0, 15)
verify(9, 1, 1, 545)

##

function measurebasin(mappe, basinbottom)
    q = [basinbottom-CartesianIndex((1, 1))]

    while !isempty(q)
        c = pop!(q)

        if mappe[c] != 0
            continue
        end

        mappe[c] = 2

        for offset ∈ [(0, 1), (0, -1), (1, 0), (-1, 0)]
            idx = c + CartesianIndex(offset)
            if checkbounds(Bool, mappe, idx) && mappe[idx] == 0
                push!(q, idx)
            end
        end
    end

    sum(mappe .== 2)
end

##

function task1(file)
    mappe = parseinput(file)
    basinbottoms = findlocmin(mappe)

    ridgemap = convert(Matrix{Int}, mappe[2:(end-1), 2:(end-1)] .== 9)

    basinsizes = map(basin -> measurebasin(copy(ridgemap), basin), basinbottoms) |> sort

    prod(basinsizes[(end-2):end])
end

##

verify(9, 2, 0, 1134)
verify(9, 2, 1, 950600)
